// -*-c++-*-
#ifndef SAOTRACE_IO_FITS_H
#define SAOTRACE_IO_FITS_H

#include <string>
#include <map>
#include <vector>
#include <iostream>
#include <algorithm>
#include <fitsio.h>

#include <Exception/Exception.h>


namespace SAOTrace {

    namespace IO {

	namespace FITS {


	    class Exception : public ::Exception {

	    public:
		Exception( const char* msg, const string& file, int status = 0 );

	    };

	    typedef std::vector<long> Extent;


	    struct ColumnBase {

		virtual void insert( void* fitsdata, long nvalues ) = 0;
		virtual int type() = 0;
		virtual void reserve( long n ) = 0;

		ColumnBase( std::string& name ) : name(name) {}
		ColumnBase( const char*  name ) : name(name) {}
		ColumnBase( std::string& name, Extent extent ) : name(name), extent(extent) {}
		long repeat;
		long naxis;
		std::string name;
		Extent extent;
	    };

	    template <typename T>
	    struct Column : ColumnBase {

		Column() {}

	        Column( std::string& name, std::vector<T>& data ) : ColumnBase( name ), data( data ) {}
	        Column( std::string& name, std::vector<T>& data, std::vector<ColumnBase*>& columns )
		    : ColumnBase( name ), data( data ) {

		    columns.push_back( this );
		}

	        Column( const char* name, std::vector<T>& data ) : ColumnBase( name ), data( data ) {}
	        Column( const char* name, std::vector<T>& data, std::vector<ColumnBase*>& columns )
		    : ColumnBase( name ), data( data ) {

		    columns.push_back( this );
		}

	        Column( std::string& name, std::vector<T>& data, Extent extent ) : ColumnBase( name, extent ), data( data ) {}
	        Column( std::string& name, std::vector<T>& data, Extent extent, std::vector<ColumnBase*>& columns )
		    : ColumnBase( name, extent ), data( data ) {
		    columns.push_back( this );
		}

		int type () { return _type ; }

		std::vector<T>& data;

		void reserve( long n ) { data.reserve(n); }

		void insert( void* fitsdata, long nvalues ) {

		    T* pdata = static_cast<T*>(fitsdata) + 1;

		    data.insert( data.end(), pdata, pdata + nvalues ) ;

		}

	    private:
		const static int _type;

	    };


	    class TableFH {

	    public:
		TableFH ( const string& file, int iomode = READONLY );

		~TableFH() {

		    if ( fits_close_file( _fptr, &_status ) )
			throw( Exception( "error closing table", _file, _status ) );
		}

		long read_table_columns( vector<FITS::ColumnBase*>& columns );
		bool has_column( const char* colname );

	    private:
		string _file;
		int _status;
		fitsfile* _fptr;

	    };


	    bool cmp_shape( std::vector<long>& a0, std::vector<long>& b0 );

	    long read_table_columns(const std::string& file, std::vector<ColumnBase*>& columns);

	}

    }

}

#endif //! SAOTRACE_IO_FITS_H
