#include <sstream>

#include "fits.h"

using namespace std;

namespace FITS = SAOTrace::IO::FITS ;


static
int read_chunk( long totaln, long offset, long firstn, long nvalues,
		int narrays, iteratorCol *data,  void *userPointer ) {

    vector< FITS::ColumnBase* >* pcolumns = static_cast< vector< FITS::ColumnBase* >* > ( userPointer );

    for( int i = 0 ; i < narrays ; i++ ) {

	FITS::ColumnBase* pcolumn = pcolumns->at(i);

	pcolumn->insert( data[i].array, nvalues * pcolumn->repeat );
    }

    return 0;
}

FITS::TableFH::TableFH( const string& file, int iomode ) : _file( file ) {

    _status = 0;

    if ( fits_open_table( &_fptr, _file.c_str(), iomode, &_status ) )
	throw( FITS::Exception( "error opening file",  _file, _status ) );

}

bool
FITS::TableFH::has_column( const char* colname ) {

    int colnum;

    // save incoming status flag
    int old_status = _status;

    fits_get_colnum( _fptr, CASEINSEN, const_cast<char*>(colname), &colnum, &_status );

    int new_status = _status;

    // reset status flag if the return from fits_get_colnum is not catastrophic
    if ( _status == COL_NOT_UNIQUE || _status == COL_NOT_FOUND )
	_status = old_status;

    return new_status != COL_NOT_FOUND;
}

long
FITS::TableFH::read_table_columns( vector<FITS::ColumnBase*>& columns ) {

    long nrows;


    if ( fits_get_num_rows( _fptr, &nrows, &_status ) )
	throw( Exception( "error getting number of rows", _file, _status ) );

    vector< iteratorCol > workcols(columns.size());
    vector< iteratorCol >::iterator workcol = workcols.begin();

    for( vector<ColumnBase*>::iterator columnp = columns.begin() ; columnp != columns.end() ; columnp++, workcol++ ) {

	ColumnBase* column = *columnp;

	workcol->fptr = _fptr;

	if ( fits_get_colnum( _fptr, CASEINSEN, const_cast<char*>( column->name.c_str() ),
			      &( workcol->colnum ), &_status ) )
	    throw( Exception( (column->name + ": not found" ).c_str(), _file, _status ) );

	workcol->datatype = column->type();
	workcol->iotype = InputCol;

	{
	    int typecode;
	    long repeat, width;
	    if ( fits_get_coltype( _fptr, workcol->colnum, &typecode, &(column->repeat), &width, &_status ) )
		throw( Exception( (column->name + string( ": error getting coltype information" )).c_str(),
				  _file, _status ) );

	    column->reserve( nrows * column->repeat);
	}

	{
	    // if column->extent isn't empty, then the caller is
	    // expecting a particular shape to a cell, and we should
	    // verify that.  otherwise, just fill in the shape.

	    // we always need to get the dimensionality of the cell
	    int naxis;

	    // need at least one viable element to take an address
	    // of. or do we?
	    Extent extent(1);

	    if ( fits_read_tdim( _fptr, workcol->colnum, 0, &naxis, &(column->extent[0]), &_status ) )
		throw( Exception( ( column->name + string( ": error getting tdim information") ).c_str(), _file, _status ) );

	    // column does *not* have a requested shape. just load the
	    // new one
	    if ( column->extent.empty() ) {

		column->extent.reserve( naxis );
		column->extent.assign( naxis, 0 );

		if ( fits_read_tdim( _fptr, workcol->colnum,
				     naxis, &naxis, &(column->extent[0]),
				     &_status )
		     )
		    throw( Exception( ( column->name + string( ": error getting tdim information")).c_str(), _file, _status ) );

	    }


	    // column has a requested shape
	    else {

		extent.reserve( naxis );
		extent.assign( naxis, 0 );

		if ( fits_read_tdim( _fptr, workcol->colnum, naxis, &naxis, &(extent[0]), &_status ) )
		    throw( ( column->name
					     + string( ": error getting tdim information in'")
					     + _file
					     + "'" )
			   );

		if ( ! cmp_shape( column->extent, extent ) ) {
		    std::ostringstream error;
		    error << column->name
			  << ": file cell shape differs from that requested "
			  << " in '" << _file << "'"
			;
		    throw( ( error.str() )  );
		}

	    }
	}
    }


    if ( fits_iterate_data( (int) workcols.size(), &(workcols[0]), 0, 0, read_chunk, &columns, &_status ) )
	throw( Exception( "error reading data", _file, _status ) );


    return nrows;
}

