#include <stdlib.h>
#include <string>
#include <algorithm>

using namespace std;

#include <gtest/gtest.h>

#include <Exception/Exception.h>

#include <saotrace/io/fits/fits.h>

namespace FITS = SAOTrace::IO::FITS;

#include "args.h"

#include "column_fits.h"

using namespace FITSTest;

// ------------------------------------------------------------------------ //

// command line arguments

class Args : public TestArgsBase {

public:

    TCLAP::ValueArg<std::string> fitsfile;

    Args( ) :
	TestArgsBase( "read_columns" ),
	fitsfile( "f", "fitsfile", "FITS file to read", true, "", "filename", cmd ) {}

};

static Args args;

// ------------------------------------------------------------------------ //

// Meta information about the columns in the input FITS file

static Column::MetaData column_meta;
static Column::MetaMap column_meta_map;


// ------------------------------------------------------------------------ //

class SingleColumnTest : public ::testing::TestWithParam<Column> {};

TEST_P( SingleColumnTest, A ) {

    Column col = GetParam();
    vector<long> data;

    vector<FITS::ColumnBase*> columns;
    FITS::Column<long> fcol( col.name, data, columns );

    long nrows = FITS::read_table_columns( args.fitsfile.getValue(), columns );

    ASSERT_EQ( nrows, data.size() / fcol.repeat  );

    ASSERT_EQ( col.repeat, fcol.repeat );
    ASSERT_EQ( col.extent, fcol.extent );

    ASSERT_EQ( col.minval, data[0] );
    ASSERT_EQ( col.maxval, data[data.size()-1] );
}

INSTANTIATE_TEST_CASE_P( Columns, SingleColumnTest, ::testing::ValuesIn( column_meta ) );


// ------------------------------------------------------------------------ //

// read in two columns; until googletest documents parameter
// generators, hard code one of them.

class DoubleColumnTest : public ::testing::TestWithParam<Column> {};

TEST_P( DoubleColumnTest, Vlong ) {

    vector<FITS::ColumnBase*> columns;

    vector<long> data[2];

    FITS::Column<long> vcol( "vlong", data[0], columns );

    Column col = GetParam();
    FITS::Column<long> fcol( col.name, data[1], columns );

    long nrows = FITS::read_table_columns( args.fitsfile.getValue(), columns );

    ASSERT_EQ( nrows, data[1].size() / fcol.repeat  );

    ASSERT_EQ( col.extent, fcol.extent );

    ASSERT_EQ( col.minval, data[1][0] );
    ASSERT_EQ( col.maxval, data[1][data[1].size()-1] );
}

INSTANTIATE_TEST_CASE_P( Columns, DoubleColumnTest, ::testing::ValuesIn( column_meta ) );

// ------------------------------------------------------------------------ //

// read in all of the columns, with extents

class AllColumnTest : public ::testing::TestWithParam<int> {

protected:

    void SetUp () {

	vdata.resize( column_meta.size() );

	for( Column::MetaData::iterator column = column_meta.begin() ; column != column_meta.end() ; ++column ) {

	    long idx = column - column_meta.begin();

	    fits_columns.push_back( new FITS::Column<long>( column->name, vdata[idx], column->extent, columns ) );
	}

	nrows = FITS::read_table_columns( args.fitsfile.getValue(), columns );

    }

    void TearDown () {

	vector< FITS::Column<long>* >::iterator fcol;

	for ( fcol = fits_columns.begin() ; fcol != fits_columns.end() ; ++fcol )
	    delete *fcol;

    }


    vector<FITS::ColumnBase*> columns;
    vector< vector<long> > vdata;
    vector< FITS::Column<long>* > fits_columns;
    long nrows;

};


TEST_P( AllColumnTest, A ) {

    int idx = GetParam();

    Column& col              = column_meta[idx];
    FITS::Column<long>* fcol = fits_columns[idx];
    vector<long>& data       = vdata[idx];

    ASSERT_EQ( nrows, data.size() / fcol->repeat  );
    ASSERT_EQ( col.repeat, fcol->repeat );
    ASSERT_EQ( col.extent, fcol->extent );
    ASSERT_EQ( col.minval, data[0] );
    ASSERT_EQ( col.maxval, data[data.size()-1] );
}

INSTANTIATE_TEST_CASE_P( Columns, AllColumnTest, ::testing::Range( 0, static_cast<int>(column_meta.size()) ) );

// ------------------------------------------------------------------------ //


// make sure that things other than <long> work
// test, no extent

template <typename T>
class TypeTest : public ::testing::Test {

public:

  typedef std::list<T> List;
  static T shared_;
  T value_;

};

typedef ::testing::Types< short, int, long, float, double > MyTypes;

TYPED_TEST_CASE(TypeTest, MyTypes);

TYPED_TEST(TypeTest, TestTypes) {

    vector<TypeParam> data;

    vector<FITS::ColumnBase*> columns;
    FITS::Column<TypeParam> fcol( "vshort", data, columns );

    long nrows = FITS::read_table_columns( args.fitsfile.getValue(), columns );

    ASSERT_EQ( nrows, data.size() / fcol.repeat  );

    Column *col = column_meta_map["vshort"];

    ASSERT_EQ( col->repeat, fcol.repeat );
    ASSERT_EQ( col->extent, fcol.extent );
    ASSERT_EQ( col->minval, data[0] );
    ASSERT_EQ( col->maxval, data[data.size()-1] );

}


// ------------------------------------------------------------------------ //

GTEST_API_ int main(int argc, char **argv) {

  args.parse( &argc, argv );

  Column::read_meta( args.fitsfile.getValue(), column_meta, column_meta_map);

  testing::InitGoogleTest(&argc, argv);

  return RUN_ALL_TESTS();
}
